'use strict';

const uuidV4 = require('uuid/v4');
const https = require('https');
const encryptor = require('simple-encryptor');

// services
const caesarEvents = require('./services/caesar-events');
const StatsDb = require('./services/caesar-stats-db');

// plugins
const eventOptionsPlugin = require('./plugins/attach-options-events-plugin');
const loggerPlugin = require('./plugins/logger-plugin');
const userRoutesPlugin = require('./plugins/get-user-routes-plugin');
const StatsDbPlugin = require('./plugins/stats-db-plugin');
const LimitRequestsPlugin = require('./plugins/limit-requests-plugin');
const errorHandlerPlugin = require('./plugins/error-handler-plugin');

// lodash
const isEmpty = require('lodash/isEmpty');
const isObject = require('lodash/isObject');
const has = require('lodash/has');

//express
const express = require('express');
const bodyParser = require('body-parser');
const helmet = require('helmet');

// middleware
const dataDecryptor = require('./middleware/req-decrypt-data');
const encryptorJson = require('./middleware/res-encryptor-json');

// logger
const pino = require('pino')();

class CaesarServer
{
    constructor (port, opts = {})
    {
        this.port = port;

        this.isSecure = false;
        if(isObject(opts) && isEmpty(opts) == false)
        {
            this.cert = opts.cert;
            this.key = opts.key;
            this.isSecure = (this.cert && this.key) ? true : false;
        }

        this.userRoutes = [];
        this.pluginRegistry = [];
        this.opts = Object.assign(opts, {port});
        this.caesarEvents = new caesarEvents();
        this.encryptor = encryptor(opts.encryptorKey);

        this.app = express();

        this.app.use( helmet() );
        this.app.use( bodyParser.json() );
        this.installPlugin( loggerPlugin() );
        this.installPlugin( eventOptionsPlugin(this.opts, this.caesarEvents) );
        this.installPlugin( StatsDbPlugin( new StatsDb() ) );
        if(opts.maxNumberOfRequests)
        {
            this.installPlugin( LimitRequestsPlugin(opts.maxNumberOfRequests) );
        }
       
        this.app.use( encryptorJson( this.encryptor ) );
        this.app.use( dataDecryptor( this.encryptor ) );

        return this;
    }

    installPlugin (plugin)
    {
        let self = this;
        let app = self.app;
        
        if(
            plugin.id != undefined &&
            plugin.name != undefined &&
            plugin.author != undefined &&
            has(plugin, 'server.routes') == true &&
            Array.isArray(plugin.server.routes) == true &&
            has(plugin, 'server.middlewares') == true &&
            Array.isArray(plugin.server.middlewares) == true
        )
        {
            plugin.server.middlewares.forEach( (middlewareCb) => {
                app.use(middlewareCb);
            });
            plugin.server.routes.forEach( (route) => {
                app.get(route.url, route.expressRouteCb);
            });
            plugin.server.routesWithPatternObject.forEach( (route) => {
                self.add(route.patternObject, route.expressRouteCb);
            });
            self.pluginRegistry.push(
                {
                    name: plugin.name,
                    author: plugin.author,
                    v: plugin.v,
                    id: plugin.id
                }
            );
        }else{
            throw new Error('Plugin is invalid!' + has(plugin, 'server.routes'));
        }
        return self;
    }

    add (patternObject, callBack, inputObject = {})
    {
        let urlId = uuidV4();
        this.userRoutes.push({
            urlId,
            patternObject,
            callBack,
            inputObject
        });
        return this;
    }

    listen () {

        let self = this;

        const port = self.port;
        self.installPlugin( userRoutesPlugin(self.userRoutes) );
        self.installPlugin( errorHandlerPlugin(self.userRoutes) );

        if(self.isSecure)
        {
            const serverOptions = {
                key: self.key,
                cert: self.cert,
                requestCert: true,
                rejectUnauthorized: false
            };
            const server = https.createServer(serverOptions, self.app).listen(port, function () {
                pino.info('Caesar Server is up and running SECURELY from port:', port, self.pluginRegistry);
                pino.info('Listening on https://' + server.address().address + ':' + port);
            });
        }else{
            self.app.listen(port, function () {
                pino.info('Caesar Server is up and running from port:', port, self.pluginRegistry);
            });
        }
    }
}

module.exports = CaesarServer;
